/* 
   Add any default pref values we want for smime
*/

pref("mail.identity.default.encryption_cert_name","");
pref("mail.identity.default.encrypt_mail_if_possible",  false);
pref("mail.identity.default.encrypt_mail_always", false);
pref("mail.identity.default.signing_cert_name", "");
pref("mail.identity.default.sign_mail", false);


